/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
# include <sys/stat.h>
# include <pwd.h>
#endif /* WIN32 */

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>

#ifndef SPTH_CL
# include <QProgressDialog>
# include <QFileInfo>
# include <QDateTime>
#endif /* SPTH_CL */

#include "dict.h"
#include "util.h"
#include "game.h"

#define DICTLST_IDXOFF_START(sign, len) \
		(dictLst_.idxOff[(len)-4][(sign)[0]-'a'][(sign)[1]-'a'][(sign)[2]-'a'].start)
#define DICTLST_IDXOFF_END(sign, len) \
		(dictLst_.idxOff[(len)-4][(sign)[0]-'a'][(sign)[1]-'a'][(sign)[2]-'a'].end)
#define NONPLURALST_IDXOFF_START(word, len) \
		(nonPluralLst_.idxOff[(len)-4][(word)[0]-'a'][(word)[1]-'a'].start)
#define NONPLURALST_IDXOFF_END(word, len) \
		(nonPluralLst_.idxOff[(len)-4][(word)[0]-'a'][(word)[1]-'a'].end)
#define WORDLST_IDXOFF_START(word, len) \
		(wordLst_.idxOff[(len)-3][(word)[0]-'a'][(word)[1]-'a'].start)
#define WORDLST_IDXOFF_END(word, len) \
		(wordLst_.idxOff[(len)-3][(word)[0]-'a'][(word)[1]-'a'].end)

namespace Spellathon
{
	static int dictSortcmp(const void *p1, const void *p2);
	static int precompSortcmp(const void *p1, const void *p2);
	static int wordSortcmp(const void *p1, const void *p2);
	static int nonPluralSortcmp(const void *p1, const void *p2);

	void
	Dict::dictOpen(const char *path)
	{
		char	line[MAX_LNLEN], *r; 
		int		slen;
		FILE	*fd;
	
		fd = fopen(path, "r");
		if (fd == 0) 
			spth_error(1, "Cannot open dictionary: '%s'", path);
	
		wordLst_.size = 0;
		int sz = 1<<14;
	
		wordLst_.arr = (WordEnt*)memalloc(sz * sizeof(WordEnt));
		memset(wordLst_.arr, 0, sz * sizeof(WordEnt));
	
again:
		while ( (r = fgets(line, MAX_LNLEN, fd)) != NULL) {
			slen = strlen(line);

			if (line[slen-1] != '\n') {
				fclose(fd);
				destroy();
				errno = EINVAL;
				spth_error(0, "'%s': Unrecognized file type. Dictionary must "
						   "be an ascii text file with the words seperated by newlines."
						   , path);
			}

			/* Remove newline char */
			line[--slen] = 0;

			/* Remove /r if file is in Windows format */
			if (line[slen-1] == '\r')
				line[--slen] = 0;
	
			if (slen < MIN_LETRS-1 || slen > MAX_LETRS)
				goto again;
	
			for (int i = 0; i < slen; i++) 
				if(!islower(line[i]))
					goto again;

			if (wordLst_.size >= sz) {
				wordLst_.arr = (WordEnt*)memresize(wordLst_.arr, sz * sizeof(WordEnt) * 2);
				memset(wordLst_.arr + sz, 0, sz * sizeof(WordEnt));
				sz *= 2;
			}
			strcpy(wordLst_.arr[wordLst_.size].word, line);
			wordLst_.arr[wordLst_.size++].len = slen;
		}
	
		if (ferror(fd)) {
			fclose(fd);
			destroy();
			spth_error(1, "'%s': Read error", path);
		}
	
		fclose(fd);

		qsort((void *)wordLst_.arr, wordLst_.size, sizeof(WordEnt), wordSortcmp);

		/* find wordLst_.arr[] offsets */
		int curlen = 0;
		memset(wordLst_.lenOff, -1, sizeof(wordLst_.lenOff)); 
		memset(wordLst_.idxOff, -1, sizeof(wordLst_.idxOff));
		char curwordidx[2] = {0, 0};
		for (int i = 0; i < wordLst_.size; i++) {
			int len = wordLst_.arr[i].len;
			if (len > curlen) {
				wordLst_.lenOff[len-3].start = i;
				curlen = len;
				curwordidx[0] = curwordidx[1] = 0;
			}
			wordLst_.lenOff[curlen-3].end = i;

			if (wordLst_.arr[i].word[0] != curwordidx[0] || 
				wordLst_.arr[i].word[1] != curwordidx[1]) 
			{
				curwordidx[0] = wordLst_.arr[i].word[0];
				curwordidx[1] = wordLst_.arr[i].word[1];
				WORDLST_IDXOFF_START(curwordidx, curlen) = i;
			}
			WORDLST_IDXOFF_END(curwordidx, curlen) = i;
		}

		dictProcess();
	}

	void
	Dict::dictProcess()
	{
		int L3sz = wordLst_.lenOff[0].end + 1;
		dictLst_.size = wordLst_.size - L3sz;

		dictLst_.arr = (DictEnt *)memalloc(dictLst_.size * sizeof(DictEnt));
		memset(dictLst_.arr, 0, dictLst_.size * sizeof(DictEnt));

		char tword[MAX_LETRS+1];
		for (int i = 0; i <  dictLst_.size; i++) {
				dictLst_.arr[i].wordp = &wordLst_.arr[i+L3sz];
				strcpy7(tword, dictLst_.arr[i].wordp->word);
				sign(tword, dictLst_.arr[i].wordp->len);
				strcpy7(dictLst_.arr[i].sign, tword);
		}

		qsort((void*)dictLst_.arr, dictLst_.size, sizeof(DictEnt), 
			  dictSortcmp);

		memset(dictLst_.idxOff, -1, sizeof(dictLst_.idxOff));
		memset(dictLst_.lenOff, -1, sizeof(dictLst_.lenOff));
		/* find dictLst_.arr[] offsets */
		int curlen = 0;
		char cursignidx[3] = { 0, 0, 0 };
		for (int i = 0; i < dictLst_.size; i++) {
			char *sign = dictLst_.arr[i].sign;
			int len = dictLst_.arr[i].wordp->len;
			if (len > curlen) {
				dictLst_.lenOff[len-4].start = i;
				curlen = len;
				cursignidx[0] = cursignidx[1] = cursignidx[2] = 0;
			}
			dictLst_.lenOff[len-4].end = i;
			if (sign[0] != cursignidx[0] || 
				sign[1] != cursignidx[1] ||
				sign[2] != cursignidx[2]) 
			{
				cursignidx[0] = sign[0];
				cursignidx[1] = sign[1];
				cursignidx[2] = sign[2];
				DICTLST_IDXOFF_START(cursignidx, curlen) = i;
			}
			DICTLST_IDXOFF_END(cursignidx, curlen) = i;
		}

		if (pluralSieve_) 
			processNonPlural();
	}

	void
	Dict::processNonPlural()
	{
		int sz = 1<<10;
		nonPluralLst_.size = 0;
		char word[MAX_LETRS+1];

		nonPluralLst_.arr = (WordEnt**)memalloc(sz * sizeof(WordEnt*));

		for (int i = wordLst_.lenOff[1].start; i < wordLst_.size; i++) {
			strcpy7(word, wordLst_.arr[i].word);
			int slen = wordLst_.arr[i].len;
			/* .*s */
			if (word[--slen] == 's') {
				word[slen] = 0;	
				if (matchWord(word, slen)) 
					continue;
				/* .*es */
				if (word[--slen] == 'e') {
					word[slen] = 0;
					if (matchWord(word, slen)) 
						continue;
				}
				/* .*ies */
				if (word[--slen] == 'i') {
					word[slen] = 'y';
					if (matchWord(word, slen+1)) 
						continue;
				}
				if (nonPluralLst_.size >= sz) {
					sz *= 2;
					nonPluralLst_.arr = (WordEnt**)memresize(nonPluralLst_.arr, sz * sizeof(WordEnt*));
				}
				nonPluralLst_.arr[nonPluralLst_.size++] = &wordLst_.arr[i];
			}
		}

		qsort((void *)nonPluralLst_.arr, nonPluralLst_.size, sizeof(WordEnt*), nonPluralSortcmp);

		memset(nonPluralLst_.idxOff, -1, sizeof(nonPluralLst_.idxOff));
		memset(nonPluralLst_.lenOff, -1, sizeof(nonPluralLst_.lenOff));
		/* find nonPluralLst_.arr[] offsets */
		char curwordidx[2] = {0, 0};
		int len, curlen = 0;
		for (int i = 0; i < nonPluralLst_.size; i++) {
			len = nonPluralLst_.arr[i]->len;
			if (len > curlen) {
				nonPluralLst_.lenOff[len-4].start = i;
				curlen = len;
				curwordidx[0] = curwordidx[1] = 0;
			}
			nonPluralLst_.lenOff[curlen-4].end = i;

			if (nonPluralLst_.arr[i]->word[0] != curwordidx[0] || 
				nonPluralLst_.arr[i]->word[1] != curwordidx[1]) 
			{
				curwordidx[0] = nonPluralLst_.arr[i]->word[0];
				curwordidx[1] = nonPluralLst_.arr[i]->word[1];
				NONPLURALST_IDXOFF_START(curwordidx, curlen) = i;
			}
			NONPLURALST_IDXOFF_END(curwordidx, curlen) = i;
		}
	}

#ifdef SPTH_CL
	Dict::Dict() 
#else
	Dict::Dict(QWidget* p) 
		: gui_h_(p)
#endif /* SPTH_CL */
	{ 
		dictbinPath_[0] = '\0';
		dictLst_.arr = 0;
		wordLst_.arr = 0;
		precompLst_.arr = 0;
		nonPluralLst_.arr = 0;
	}

	void
	Dict::init(const char *path, bool precomp, bool pluralSieve) 
	{
		dictPath_ = path;
		destroy();
		precomp_ = precomp;
		pluralSieve_ = pluralSieve;
		if (precomp_) {
			setPrecompPath();
			precompOpen(path); 
		} else {
			dictOpen(path); 
		}
	}
	
	void
	Dict::setPrecompPath()
	{
#ifdef WIN32
		strcpy(dictbinPath_, getenv("APPDATA"));
		strcat(dictbinPath_, "/"DICTBIN);
#else
		struct passwd *pw = getpwuid(getuid());
		if (pw == 0)
			spth_error(1, "Cannot get user info.");

		strcpy(dictbinPath_, pw->pw_dir);
		strcat(dictbinPath_, "/.spellathon/");
		if (access(dictbinPath_, R_OK | W_OK) < 0) {
			if (mkdir(dictbinPath_, 
			          S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
 			       spth_error(1, "Cannot create '%s'", dictbinPath_);
		}
			
		strcat(dictbinPath_, DICTBIN);
#endif /* WIN32 */
	}

	void
	Dict::precompOpen(const char *path)
	{
		FILE *binfl = fopen(dictbinPath_, "rb");	
		if (binfl == 0) {
			if (errno == ENOENT) {
				precompCreate(path);
				destroy();
				precompOpen(path);
				return;
			} else {
 			       spth_error(1, "Cannot open precomputed dictionary: '%s'", dictbinPath_);
			}
		}
			

		if (fread(&precompHdr_, sizeof(precompHdr_), 1, binfl) < 
			(unsigned int)1) {
			goto oops;
		}

#ifdef SPTH_CL 
# ifndef WIN32
		struct stat binp, dictp;
		stat(dictbinPath_, &binp);
		stat(path, &dictp);
# endif
#endif /* SPTH_CL */
		if (strcmp(precompHdr_.dictPath, path) ||
		    strcmp(precompHdr_.magic, MAGIC) ||
			precompHdr_.version != PRECOMP_VER ||
			(precompHdr_.pluralSieve > 0) != pluralSieve_
#ifdef SPTH_CL 
# ifndef WIN32
		    || binp.st_mtime < dictp.st_mtime
# endif
#else
		    || QFileInfo(dictbinPath_).lastModified().toTime_t() <
		    QFileInfo(path).lastModified().toTime_t()
#endif /* SPTH_CL */
		) {
			fclose(binfl);
			precompCreate(path);
			destroy();
			precompOpen(path);
			return;
		}

		memcpy(wordLst_.lenOff, precompHdr_.wordLst_lenOff, sizeof(wordLst_.lenOff));
		memcpy(wordLst_.idxOff, precompHdr_.wordLst_idxOff, sizeof(wordLst_.idxOff));

		wordLst_.size = precompHdr_.wordLst_size;
		wordLst_.arr = (WordEnt*)memalloc(wordLst_.size * sizeof(WordEnt));
		if (fread(wordLst_.arr, sizeof(WordEnt), wordLst_.size, binfl) < (unsigned int)wordLst_.size) 
			goto oops;

		precompLst_.size = precompHdr_.precompLst_size;
		precompLst_.arr = (PrecompEnt*)memalloc(precompLst_.size * sizeof(PrecompEnt));
		if (fread(precompLst_.arr, sizeof(PrecompEnt), precompHdr_.precompLst_size,
				  binfl) 
			< (unsigned int)precompHdr_.precompLst_size) 
			goto oops;

		dictProcess();

		fclose(binfl);
		return;

oops:
		fclose(binfl);
		destroy();
		spth_error(1, "'%s': Read error", dictbinPath_);
	}

	void
	Dict::precompCreate(const char *path)
	{
		int cnt = 0, n;

		dictOpen(path);
		Game	game(this, pluralSieve_);

		precompLst_.arr = (PrecompEnt*)memalloc(nent7() * 7 * sizeof(PrecompEnt));

#ifndef SPTH_CL
		QProgressDialog pdiag("Precomputing Dictionary ...", 0, 
							  dictLst_.lenOff[7-4].start , dictLst_.lenOff[7-4].end, gui_h_);
		pdiag.setWindowModality(Qt::WindowModal);
		pdiag.setMinimumDuration(0);
		pdiag.setWindowFlags(Qt::Popup);
		pdiag.move((gui_h_->size().width()-pdiag.size().width())/2 + gui_h_->pos().x(),
				   gui_h_->size().height()/2 + gui_h_->pos().y());
#endif /* SPTH_CL */

		if (dictLst_.lenOff[7-4].start != -1) {
			for (int i = dictLst_.lenOff[7-4].start; 
				 i <= dictLst_.lenOff[7-4].end; i++) {
				char sig[MAX_LETRS+1];
				strcpy7(sig, dictLst_.arr[i].sign);
	
				char cltried[MAX_LETRS];
				int k = MAX_LETRS - 2;
				int clidx = 0;
				memset(cltried, 0, sizeof(cltried));
				do {
					if ((n = game.solver(sig)) > 0) {
						strcpy7(precompLst_.arr[cnt].sign, sig);	
						precompLst_.arr[cnt++].nsol = n;
					}

					/* find non-duplicate central letter */
					cltried[clidx++] = sig[MAX_LETRS-1];
again:
					for (char *p = cltried; *p && k >= 0; p++) {
						if (*p == sig[k]) {
							k--;
							goto again;
						} 
					}

					if (k < 0)
						break;

					/* try a different central letter */	
					swapch(&sig[MAX_LETRS-1], &sig[k--]);
				} while (1);

				if (!(i & ((1<<9) - 1))) {
#ifndef SPTH_CL
					pdiag.setValue(i);
#else
					printf("\rPrecomputing dictionary... %d%%", (i-dictLst_.lenOff[7-4].start)*100/nent7());
					fflush(stdout);
#endif /* SPTH_CL */
				}
			}
		}

#ifdef SPTH_CL
                printf("\rPrecomputing dictionary... done.\n");
#endif /* SPTH_CL */

		/* sort by number of solutions */
		qsort((void *)precompLst_.arr, cnt, sizeof(PrecompEnt), precompSortcmp);

		/* remove dup signs */
		if (cnt) {
			PrecompEnt *f = precompLst_.arr;
			PrecompEnt *l = precompLst_.arr + cnt;
			PrecompEnt *r = f;
			while (++f != l) {
				if (strcmp(r->sign, f->sign))
					memmove(++r, f, sizeof(PrecompEnt));
			}
			++r;
			cnt = r - precompLst_.arr;
		}

		/* fill header */
		memset(&precompHdr_, 0, sizeof(precompHdr_));
		precompHdr_.wordLst_size = wordLst_.size;
		precompHdr_.precompLst_size = cnt;
		precompHdr_.version = PRECOMP_VER;
		strncpy(precompHdr_.dictPath, path, MAX_PATH_LEN);
		precompHdr_.dictPath[MAX_PATH_LEN-1] = 0;
		strcpy(precompHdr_.magic, MAGIC);
		precompHdr_.pluralSieve = pluralSieve_;
		memcpy(precompHdr_.wordLst_lenOff, wordLst_.lenOff, sizeof(wordLst_.lenOff));
		memcpy(precompHdr_.wordLst_idxOff, wordLst_.idxOff, sizeof(wordLst_.idxOff));
		
		FILE *binfl = fopen(dictbinPath_, "wb");
		if (binfl == 0) {
			destroy();
			spth_error(1, "Cannot open precomputed dictionary: '%s'", dictbinPath_);
		}

		if (fwrite(&precompHdr_, sizeof(precompHdr_), 1, binfl) < 1 ||
			fwrite(wordLst_.arr, sizeof(WordEnt), wordLst_.size, binfl) < (unsigned int)wordLst_.size ||
			fwrite(precompLst_.arr, sizeof(PrecompEnt), cnt, binfl) < (unsigned int)cnt
		   ) {
			fclose(binfl);
			destroy();
			spth_error(1, "'%s': Write error", dictbinPath_);
		}

		fclose(binfl);
	}

	/* binary search the dict. */
	/* performance critical routine */
	inline int
	Dict::bsearch(const char * sign, int slen) const
	{
		int i, l, h, m, r;
	
		/* slen is >= 4 */
		l = DICTLST_IDXOFF_START(sign, slen);
		h = DICTLST_IDXOFF_END(sign, slen);
		
		if (l + h >= 0) {
			char *s;
			while (h >= l) {
				m = (l + h) / 2;
	
				/* r = strcmp(sign, dictLst_.arr[m].sign) */
				r = 0;
				i = 0;
				s = dictLst_.arr[m].sign; 
				if (sign[0] != s[0]) { goto out; }
				if (sign[1] != s[1]) { i = 1; goto out; }
				if (sign[2] != s[2]) { i = 2; goto out; }
				if (sign[3] != s[3]) { i = 3; goto out; }
				if (sign[4] != s[4]) { i = 4; goto out; }
				if (sign[5] != s[5]) { i = 5; goto out; }
				if (sign[6] != s[6]) { i = 6; }
out:
				r = sign[i] - s[i];
	
				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
						return m;
			}	
		}

		return -1;
	}

	inline bool
	Dict::isNotPlural(const char *word, int slen) const
	{
		int i, l, h, m, r;
	
		if (word[slen-1] != 's')
			return true;

		l = NONPLURALST_IDXOFF_START(word, slen);
		h = NONPLURALST_IDXOFF_END(word, slen);
		
		if (l + h >= 0) {
			char *pl;
			while (h >= l) {
				m = (l + h) / 2;
	
				r = 0;
				i = 0;
				pl = nonPluralLst_.arr[m]->word;
				if (word[0] != pl[0]) { goto out; }
				if (word[1] != pl[1]) { i = 1; goto out; }
				if (word[2] != pl[2]) { i = 2; goto out; }
				if (word[3] != pl[3]) { i = 3; goto out; }
				if (word[4] != pl[4]) { i = 4; goto out; }
				if (word[5] != pl[5]) { i = 5; goto out; }
				if (word[6] != pl[6]) { i = 6; }
out:
				r = word[i] - pl[i];
	
				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return true;
			}	
		}

		return false;
	}

	bool
	Dict::matchWord(const char *word, int slen) const
	{
		int i, l, h, m, r;
	
		l = WORDLST_IDXOFF_START(word, slen);
		h = WORDLST_IDXOFF_END(word, slen);
		
		if (l + h >= 0) {
			char *w;
			while (h >= l) {
				m = (l + h) / 2;
	
				r = 0;
				i = 0;
				w = wordLst_.arr[m].word;
				if (word[0] != w[0]) { goto out; }
				if (word[1] != w[1]) { i = 1; goto out; }
				if (word[2] != w[2]) { i = 2; goto out; }
				if (word[3] != w[3]) { i = 3; goto out; }
				if (word[4] != w[4]) { i = 4; goto out; }
				if (word[5] != w[5]) { i = 5; goto out; }
				if (word[6] != w[6]) { i = 6; }
out:
				r = word[i] - w[i];
	
				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return true;
			}	
		}

		return false;
	}
	
	
	void
	Dict::matchSign(const char * sign, char *w[], int *idx, int *has7) const
	{
		int pos, npos, slen;
	
		slen = strlen(sign);
	
		/* binary search dictionary for matching signature. */
		pos = bsearch(sign, slen);
	
		if (pos != -1) {
			/* 
			 * search for all remaining duplicate signatures.
			 * the duplicates can be in contiguous locations
			 * on both the left and right sides of the index 
			 * returned by bsearch(). 
			 */
			npos = pos - 1;
			int s = DICTLST_IDXOFF_START(sign, slen);
			int e = DICTLST_IDXOFF_END(sign, slen);
			/* scan left */
			while (npos >= s && strcmp7_eq(dictLst_.arr[npos].sign, sign)) {
				if (!pluralSieve_ || 
					isNotPlural(dictLst_.arr[npos].wordp->word, slen)) 
				{
					w[(*idx)++] = dictLst_.arr[npos].wordp->word;

					if (slen == 7)
						(*has7)++;

					if (*idx > SOLVER_MAX_WORDS)
						return;
				}
				npos--;
			}
			/* scan right */
			do {
				if (!pluralSieve_ || 
					isNotPlural(dictLst_.arr[pos].wordp->word, slen)) 
				{ 
					w[(*idx)++] = dictLst_.arr[pos].wordp->word;

					if (slen == 7)
						(*has7)++;

					if (*idx > SOLVER_MAX_WORDS)
						return;
				}
			} while (++pos < e && strcmp7_eq(dictLst_.arr[pos].sign, sign));
		}
	}
	
	/* sort by strlen() and then lexicographically */
	static int
	dictSortcmp(const void *p1, const void *p2)
	{
		const DictEnt *s1 = (const DictEnt*)p1;
		const DictEnt *s2 = (const DictEnt*)p2;
	
		int r = s1->wordp->len - s2->wordp->len;
		if (r == 0) 
			return strcmp(s1->sign, s2->sign);
		else 
			return r;
	}
	
	static int
	wordSortcmp(const void *p1, const void *p2)
	{
		const WordEnt *s1 = (const WordEnt*)p1;
		const WordEnt *s2 = (const WordEnt*)p2;
		
		int r = s1->len - s2->len;
		if (r == 0) 
			return strcmp(s1->word, s2->word);
		else 
			return r;
	}
	
	static int
	precompSortcmp(const void *p1, const void *p2)
	{
		const PrecompEnt *s1 = (const PrecompEnt *)p1;
		const PrecompEnt *s2 = (const PrecompEnt *)p2;
		
		int r = s1->nsol - s2->nsol;
		if (r == 0) 
			return strcmp(s1->sign, s2->sign);
		else 
			return r;
	}

	static int
	nonPluralSortcmp(const void *p1, const void *p2)
	{
		const WordEnt *s1 = *(const WordEnt **)p1;
		const WordEnt *s2 = *(const WordEnt **)p2;
		
		int r = s1->len - s2->len;
		if (r == 0) 
			return strcmp(s1->word, s2->word);
		else 
			return r;
	}
	
	void
	Dict::destroy()
	{
		free(dictLst_.arr);			dictLst_.arr = 0;
		free(wordLst_.arr);			wordLst_.arr = 0;
		free(nonPluralLst_.arr); 	nonPluralLst_.arr = 0;
		free(precompLst_.arr);		precompLst_.arr = 0;
	}
}
